#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#pragma pack (1)

typedef unsigned char uchar;
typedef enum {FALSE, TRUE} Bool;

// #define PACKET_DATA_SIZE 100

typedef struct frame_header {
	uchar sof;
	uchar dest:4;
	uchar src:4;
	uchar size;
	uchar fid;
#ifdef PACKET_DATA_SIZE
	uchar data[PACKET_DATA_SIZE];
#endif
} frame;

typedef enum errorid_types {FRAME_OK, ERROR_CRC, ERROR_DEST, ERROR_FID} errorid;

#define SOF ((uchar) 0xFF)

#define MY_ID1 (1)
#define MY_ID2 (2)
#define MY_ID3 (3)
#define HIS_ID4 (4)
#define HIS_ID5 (5)
#define HIS_ID6 (6)
#define HIS_ID7 (7)

frame *FRAME = NULL;

frame * create_frame(int to, int from, int seq, char *data);
void destroy_frame(frame *frm);

void frame_RX();
Bool process_frame(frame *rcvdFrame);

Bool valid_Frame(frame *rcvdFrame);
uchar check_CRC(void *ptr, int size);

frame * accept_frame(frame *rcvdFrame);
uchar * get_frame_data(frame *rcvdFrame);

uchar update_fid(frame *rcvdFrame);


int main()
{
	int i;
	char buffer[100];

	for (i=0; i<3; i++)
	{
		memset (buffer, 0, sizeof(buffer));

		printf("please enter a message: ");
		scanf("%s", buffer);

		FRAME = create_frame(1, 3,(i+1)%16, buffer);

		frame_RX();
		
		destroy_frame(FRAME);
		FRAME=NULL;
	}

	return 0;
}

frame * create_frame(int to, int from,int seq, char *data) // creating a frame
{
	int len;
	frame *sendFrame = NULL;
	uchar *pCRC;

#ifdef PACKET_DATA_SIZE
	len = PACKET_DATA_SIZE;
#else
	len = strlen(data) + 1; // len will include the '\0' character in case of a string ..
#endif

	sendFrame = (frame *) malloc(sizeof(frame) + len + 1); // packet should have also trailer (the CRC)
	memset(sendFrame, 0, sizeof(frame) + len + 1);

	sendFrame->sof = SOF;
	sendFrame->dest = to;
	sendFrame->src = from;
	sendFrame->size = len;
	sendFrame->fid = seq;
#ifdef PACKET_DATA_SIZE
	strcpy((char *)sendFrame->data, data);
	pCRC=(uchar *)(sendFrame+1); 
	*pCRC = check_CRC(sendFrame->data,sendFrame->size);
#else
	strcpy((char *) (sendFrame+1), data);
	pCRC = (uchar *)(sendFrame+1) + len;
	*pCRC = check_CRC(sendFrame + 1,sendFrame->size);
#endif

	return sendFrame;
}

void destroy_frame(frame *frm) // destroying a frame
{
	free(frm);
}

Bool valid_Frame(frame *rcvdFrame)  // checking if frame is valid through sof and crc
{
	uchar *pCRC = NULL;
	uchar *pData = NULL;

#ifdef PACKET_DATA_SIZE
	pData = rcvdFrame->data;
#else
	pData = (uchar *) (rcvdFrame + 1);
#endif

	pCRC= (uchar *) pData + rcvdFrame->size;

	if ( rcvdFrame->sof == SOF && *pCRC == check_CRC(pData, rcvdFrame->size) )
		return TRUE;
	else
		return FALSE;
}

// it is not recommended, two reasons:
// 1. it is much faster to copy the address insted of the whole struct
// 2. assuming the frame has a varied data length, it will be impossible to send the whole frame

frame * accept_frame(frame *rcvdFrame) // accepting frame by dst
{
	int bValid = 0;

	switch (rcvdFrame->dest)
	{
	case MY_ID1:
	case MY_ID2:
	case MY_ID3:
		bValid = 1;
		break;
	default:
		bValid = 0;
		break;
	}

	if (bValid)
		return rcvdFrame;
	else
		return NULL;
}

uchar * get_frame_data(frame *rcvdFrame)
{
#ifdef PACKET_DATA_SIZE
	return rcvdFrame->data;
#else
	return (uchar *) (rcvdFrame += 1); // go to the end of the header
#endif
}

uchar check_CRC(void *ptr, int size) // creating the crc .. sum of all the data
{
	int i;
	uchar *pData=(uchar *)ptr, CRC=0;

	for (i=0; i<size; i++, pData++)
		CRC += *pData;

	return CRC;
}

// We don't have enough. the max size in frame struct is 0xFF (255), 
// max data each is 0xFF (255), so max crc 65025, we only save the LSB of that ...

uchar update_fid(frame *rcvdFrame)
{
	static uchar SrcFid[16] = {0};

	if ( ( SrcFid[rcvdFrame->src]+1 )%16 != rcvdFrame->fid ) // checking if fid is the expected fid
	{
		SrcFid[rcvdFrame->src] = rcvdFrame->fid;  // if not, updating it and sending error
		return -1;
	}
	else
	{
		SrcFid[rcvdFrame->src]++;
		SrcFid[rcvdFrame->src]%=16;
		return SrcFid[rcvdFrame->src]; // if it's ok, sending the fid
	}
}

Bool process_frame(frame *rcvdFrame, errorid *err)
{
	uchar fid;

	if (!valid_Frame(rcvdFrame))  // checking frame's validity
	{
		*err = ERROR_CRC;
		return FALSE; // for uchar -1 is oxFF / 255
	}

	if (!accept_frame(rcvdFrame)) // accepting frame by dst
	{
		*err = ERROR_DEST;
		return FALSE; // for uchar -1 is oxFF / 255
	}

	if ( -1 == (fid = update_fid(rcvdFrame)) ) // checking and updating fid
	{
		*err = ERROR_FID;
		return FALSE; // for uchar -1 is oxFF / 255
	}

	*err = FRAME_OK;
	return TRUE;
}

void frame_RX()
{
	frame *rcvdFrame;
	static errorid DestErr[16]; // all static variables are by default 0;
	errorid err;
	
	//getting the frame from the pool ..//

	rcvdFrame = FRAME;
	process_frame(rcvdFrame, &err);

	// sometimes we will be responding according to error type ..
	// but right now we only have to save it to a database
	DestErr[rcvdFrame->dest] = err;
}

// problems using this frame implementations are:
//  - src and dst are assembled from only 4 bits, leaving only 16 possible addresses
//  - only data is going through checksum, if the error occur in the frame itself the protocol will miss it
//  - the frame is limited to only 255 bytes data length
//  - crc might be unreliable cause it's only 8 bits checksum
//  - fid (frame id) is unreliable cause it's only an 8 bits sequence